// Project: Control a 3D printer (plotter) through command line
// Homepage:
// https://homofaciens.de/technics-machines-3D-printer-Zonestar-Plotter_en.htm
// Author: Norbert Heinz
// Version: 0.1
// Creation date: 10.04.2019
// This program is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation version 3 of the License.
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// For a copy of the GNU General Public License see http://www.gnu.org/licenses/
//
// compile with:
// gcc Gcode-console.c -o Gcode-console -lm

#include <stdio.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <dirent.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>


#define BAUDRATE B115200
#define ARDUINOPORT "/dev/ttyUSB0"
//#define ARDUINOPORT "/dev/ttyACM0"
#define FALSE 0
#define TRUE 1


#define PI 3.1415927


double Scale = 1.0;
char FileName[1000] = "";

int fd = 0;
struct termios TermOpt;

char PicturePath[1000];
char allowedChars[1000];



//+++++++++++++++++++++++ Start readport +++++++++++++++++++++++++++++++
char  readport(int printBuff){
  int n;
  char buff;
  n = read(fd, &buff, 1);
  if(n > 0){
    if(printBuff == 1){
      printf("%c", buff);
    }
    return buff;
  }
  return 0;
}
//------------------------ End readport --------------------------------

//+++++++++++++++++++++++ Start sendport +++++++++++++++++++++++++++++++
void sendport(unsigned char ValueToSend){
  int n;
  int BytesInBuffer;


//  ioctl(fd,TIOCOUTQ,&BytesInBuffer);
//  while(BytesInBuffer > MAX_ARDUINO_BUFFER){//Wait until bytes in write buffer are below maximum
//    ioctl(fd,TIOCOUTQ,&BytesInBuffer);
//  }

  n = write(fd, &ValueToSend, 1);

  if (n < 0){
    printf("write() of value failed!\n");
  }
//  else{
//    arduinoBuffer++;
//    if(arduinoBuffer == MAX_ARDUINO_BUFFER){
//      while(readport(1) != 0);
//readport();
//      arduinoBuffer = 0;
//    }
//  }

}
//------------------------ End sendport --------------------------------

//+++++++++++++++++++++++ Start openport +++++++++++++++++++++++++++++++
int openport(void){
    int openRet = 0;

    fd = open(ARDUINOPORT, O_RDWR | O_NOCTTY | O_NDELAY);
    if (fd == -1)  {
      printf("init_serialport: Unable to open port\n");
      openRet = 1;
    }

    if (tcgetattr(fd, &TermOpt) < 0) {
      printf("init_serialport: Couldn't get term attributes\n");
      openRet = 1;
    }
    speed_t brate = BAUDRATE; // let you override switch below if needed

    cfsetispeed(&TermOpt, brate);
    cfsetospeed(&TermOpt, brate);

    // 8N1
    TermOpt.c_cflag &= ~PARENB;
    TermOpt.c_cflag &= ~CSTOPB;
    TermOpt.c_cflag &= ~CSIZE;
    TermOpt.c_cflag |= CS8;
    // no flow control
    TermOpt.c_cflag &= ~CRTSCTS;

    TermOpt.c_cflag |= CREAD | CLOCAL;  // turn on READ & ignore ctrl lines
    TermOpt.c_iflag &= ~(IXON | IXOFF | IXANY); // turn off s/w flow ctrl

    TermOpt.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // make raw
    TermOpt.c_oflag &= ~OPOST; // make raw

    // see: http://unixwiz.net/techtips/termios-vmin-vtime.html
    TermOpt.c_cc[VMIN]  = 0;
    TermOpt.c_cc[VTIME] = 20;

    if( tcsetattr(fd, TCSANOW, &TermOpt) < 0) {
      printf("init_serialport: Couldn't set term attributes\n");
      openRet = 1;
    }
    return (openRet);
}
//------------------------ End openport --------------------------------

//+++++++++++++++++++++++ Start sendLine +++++++++++++++++++++++++++++++
void sendLine(char *Line){
  int i;
  char c;
  char receivedLine[10000];
  int LineOK = 1;

  //Check if line is valid
  for(i=0; i<strlen(Line); i++){
    if(strchr(allowedChars, Line[i])==NULL){
      LineOK = 0;
    }
  }

  if(LineOK){
    if(strlen(Line) > 2){
      strcat(Line, "\n");
      for(i=0; i<strlen(Line); i++){
        sendport(Line[i]);
      }
    }
    while(readport(0) !=0 );//Empty read buffer
    receivedLine[0]='\0';
    i = 0;
    while(strstr(receivedLine, "wait")==NULL && strstr(receivedLine, "ok")==NULL){
      c = readport(0);
      if(c != 0 && c != '\0' && c != '\n'){
        receivedLine[i]=c;
        i++;
        receivedLine[i]='\0';
        //printf("%s\n", receivedLine);
      }
      usleep(1000);
    }
  }
  else{
    printf("Not allowd character in line!\n");
  }


}
//------------------------ End sendLine --------------------------------


//######################################################################
//################## Main ##############################################
//######################################################################

int main(int argc, char **argv){

  char FullFileName[1000] = "";
  int i;
  FILE *PlotFile;
  #define TEXT_LINE_MAX 200
  char TextLine[TEXT_LINE_MAX];
  long coordinateCount = 0;
  char a;
  int ReadState = 0;
  char commandLine[1000];


  strcpy(allowedChars, "XYZGMFxyzgmf1234567890. -");

  int endLoop = 0;
  int stopPlot = 0;


  if(openport() != 0){
    exit(1);
  }

  printf("Waiting for boot to finish...\n");
    sleep(10);//Wait until reboot of printer is finished
  printf("...done!\n");

  strcpy(FileName, "noFiLE");

  getcwd(PicturePath, 1000);
  strcat(PicturePath, "/pictures");
  //printf("PicturePath=>%s<\n", PicturePath);


  //Home all axes
  printf("Homing all axes...\n");
  strcpy(TextLine, "G1 Z10.00\r");
  sendLine(TextLine);
  strcpy(TextLine, "G28 X Y Z\r");
  sendLine(TextLine);
  sleep(15);
  printf("...done!\n");


  printf("Lifting pen...\n");
  strcpy(TextLine, "G1 Z10.00\r");
  sendLine(TextLine);
  printf("...done!\n");

  printf("Program call with %d parameters\n", argc);


  if(argc > 1){//Print gcode file
    printf("Command line argument = '%s'\n\n", argv[1]);

    for(int realRun = 0; realRun < 2; realRun++){
      if((PlotFile=fopen(argv[1],"rb"))==NULL){
        printf("Could not open or find file '%s'\n\n", argv[1]);
        return(1);
      }
      else{
        while(!(feof(PlotFile)) && stopPlot == 0){
          fread(&a, 1, 1, PlotFile);
          i=0;
          TextLine[0] = '\0';
          while(!(feof(PlotFile)) && a != 10 && i < TEXT_LINE_MAX - 2){
            if(strchr(allowedChars, a)>0){
              TextLine[i] = a;
              TextLine[i+1] = '\0';
              i++;
            }
            else{
              printf("Not allowed char'%c' in file!", a);
              exit(1);
            }
            fread(&a, 1, 1, PlotFile);
            if(a == ';'){//Discard until end of line
              while(!(feof(PlotFile)) && a != 10){
                fread(&a, 1, 1, PlotFile);
              }
            }
          }
          printf("%s\n", TextLine);
          if(realRun == 1){
            sendLine(TextLine);
          }
        }
        fclose(PlotFile);
        if(realRun == 0){
          printf("\n\n...Check OK...\n");
        }
        else{
          printf("\n\n...done!\n");
        }
      }
    }
    //sleep(10);
    close(fd);
    return(0);
  }



  while(endLoop == 0){
    printf("Type Gcode line: ");

    for(i=0; i<TEXT_LINE_MAX; i++){
      TextLine[i] = '\0';
    }
    char c = 0;
    i=0;

    while(c != '\n'){
      c = getchar();
      if(c != '\n' && i < TEXT_LINE_MAX - 2){
        TextLine[i] = c;
        i++;
        TextLine[i] = '\0';
        //printf(">%c<", c);
      }
    }


    if(strstr(TextLine, "CIRCLE") != NULL){
      double r = 50;
      double alpha = 0.0, alphaStart = PI/2.0;
      double corners = 3;
      double X=0.0, Y=0.0, Z=10.0, F=50.0 * 60.0;
      double Xoffset = 100.0, Yoffset = 100.0;
      long sleepPause = 3;

      sprintf(TextLine, "G1 Z10.0 F%.3lf", F);
      sendLine(TextLine);

      for(corners=3; corners<100; corners++){
        sleep(5);
        if(corners == 5){
          corners = 8;
        }
        if(corners == 9){
          corners = 100;
          sleepPause = 0;
        }

        if(corners == 4){
          alphaStart = PI / 4.0;
        }
        else if(corners == 8){
          alphaStart = PI / 8.0;
        }
        else{
          alphaStart = PI / 2.0;
        }

        Z=10.0;
        for(alpha=0.0+alphaStart; alpha<2.0*PI+alphaStart+(2.0*PI/corners)-0.001; alpha+=2.0 * PI / corners){
          X = cos(alpha) * r + Xoffset;
          Y = sin(alpha) * r + Yoffset;
          strcpy(TextLine, "G1");
          sprintf(TextLine, "%s X%.3lf", TextLine, X);
          sprintf(TextLine, "%s Y%.3lf", TextLine, Y);
          sprintf(TextLine, "%s Z%.3lf", TextLine, Z);
          sprintf(TextLine, "%s F%.3lf", TextLine, F);
          sendLine(TextLine);
          printf("%s", TextLine);
          sleep(sleepPause);
          if(Z==10.0){
            Z = 0.0;
            strcpy(TextLine, "G1");
            sprintf(TextLine, "%s X%.3lf", TextLine, X);
            sprintf(TextLine, "%s Y%.3lf", TextLine, Y);
            sprintf(TextLine, "%s Z%.3lf", TextLine, Z);
            sprintf(TextLine, "%s F%.3lf", TextLine, F);
            sendLine(TextLine);
            printf("%s", TextLine);
            sleep(sleepPause);
          }
        }
        Z = 10.0;
        strcpy(TextLine, "G1");
        sprintf(TextLine, "%s X%.3lf", TextLine, X);
        sprintf(TextLine, "%s Y%.3lf", TextLine, Y);
        sprintf(TextLine, "%s Z%.3lf", TextLine, Z);
        sprintf(TextLine, "%s F%.3lf", TextLine, F);
        sendLine(TextLine);
        printf("%s", TextLine);
        sleep(sleepPause);

        X = 0.0;
        Y = 0.0;
        strcpy(TextLine, "G1");
        sprintf(TextLine, "%s X%.3lf", TextLine, X);
        sprintf(TextLine, "%s Y%.3lf", TextLine, Y);
        sprintf(TextLine, "%s Z%.3lf", TextLine, Z);
        sprintf(TextLine, "%s F%.3lf", TextLine, F);
        sendLine(TextLine);
        printf("%s\n\n", TextLine);
        sleep(sleepPause);
      }
    }
    else{
      if(strlen(TextLine) == 0){
        endLoop = 1;
      }
      else{
        double X=-99.999, Y=-99.999, Z=-99.999, F=-1.0;
        for(i=0; i<strlen(TextLine); i++){
          if(toupper(TextLine[i]) == 'X'){
            X = strtod(TextLine+i+1,NULL);
          }
          if(toupper(TextLine[i]) == 'Y'){
            Y = strtod(TextLine+i+1,NULL);
          }
          if(toupper(TextLine[i]) == 'Z'){
            Z = strtod(TextLine+i+1,NULL);
          }
          if(toupper(TextLine[i]) == 'F'){
            F = strtod(TextLine+i+1,NULL);
          }
        }
        strcpy(TextLine, "G1");
        if(X > -99.999){
          sprintf(TextLine, "%s X%.3lf", TextLine, X);
        }
        if(Y > -99.999){
          sprintf(TextLine, "%s Y%.3lf", TextLine, Y);
        }
        if(Z > -99.999){
          sprintf(TextLine, "%s Z%.3lf", TextLine, Z);
        }
        if(F > 0){
          F = F *60;//Convert mm/s to mm/min (default in Repetier)
          sprintf(TextLine, "%s F%.3lf", TextLine, F);
        }
        sendLine(TextLine);
        printf(">%s< processed\n", TextLine);
      }
    }

  }


  close(fd);

  return 0;
}
