#! /usr/bin/perl

# Script converts Bitmaps into Gcode commands for plotting
# More infos on the project page:
# https://homofaciens.de/technics-machines-3D-printer-Zonestar-Printer_en.htm

use strict;
use IO::Handle;


my $i = 0;
my $j = 0;
my $line_in = "";

my $progress = 0;
my $start = 0;
my $stop = 0;
my $X=0;
my $Y=0;
my $Z=0;
my $E=0;
my $feed_rate = (200.000 * 60);
my $feed_rate_pen_change = (50.00 * 60);#feedrate in mm per minute
my $feedrate_filament = (100 * 60);

my $Z_up = 16.100;
my $Z_down = 0.501;#Z=0.000 would trigger Z switch whenever Pen is down

my $E_up = 24.323;
my $E_down = 15.501;

my $dotWidth = 5.600;
my $dotHeight = 5.000;

my $pen_mode = 1;   #0 - normal, 1 - extruder motor

my @feedrate_array;
my @remarks;
my @X_coords;
my @Y_coords;
my @Z_coords;
my @E_coords;

my $coord_no = 0;
my $dot_no = 0;


my $x_min = 99999;
my $x_max = -99999;
my $y_min = 99999;
my $y_max = -99999;

my $bed_size_x = 220;
my $bed_size_y = 220;
my $x_offset = 0;#Offset needed to center plot
my $y_offset = 0;#Offset needed to center plot

#Pen position correction relative to printer nozzle
my $pen_Y = 0;
my $pen_X = 0;

my $scale = 1.00;
my $mirror_X = 0;
my $mirror_Y = 1;

my $SourceFile = "";

my $rgb_r = 0;
my $rgb_g = 0;
my $rgb_b = 0;

my $color = 0;
my $color_new = -1;
my $color_old = -1;

my @pen_X;
my @pen_gap_X;
my @pen_Z_low;
my @pen_Z_low_empty;
my @pen_Z_high;

#Set $color_no = 1 for monochrome mode
my $color_no = 4;


#Coordinates for pen change
#normal values (Z switch spacer = 18mm):
#pen_Z_high, 203;
#pen_Z_low_empty, 90;
#values T-shirt mount (Z switch spacer = 43mm):
#pen_Z_high, 180;
#pen_Z_low_empty, 69;

#Coordinates pen 1
push @pen_X, -10;
push @pen_gap_X, 34;
push @pen_Z_low, 20;
push @pen_Z_low_empty, 90;
push @pen_Z_high, 203;

#Coordinates pen 2
push @pen_X, 79;
push @pen_gap_X, 34;
push @pen_Z_low, 20;
push @pen_Z_low_empty, 90;
push @pen_Z_high, 203;

#Coordinates pen 3
push @pen_X, 127;
push @pen_gap_X, 171;
push @pen_Z_low, 20;
push @pen_Z_low_empty, 90;
push @pen_Z_high, 203;

#Coordinates pen 4
push @pen_X, 215;
push @pen_gap_X, 171;
push @pen_Z_low, 20;
push @pen_Z_low_empty, 90;
push @pen_Z_high, 203;



my $Graphic = "";

my $FileMode = 1;

my $SourceFile = "";

my $num_args = $#ARGV + 1;

my $PI = 3.1415927;

my $longRead;
my $intRead;
my $stringRead;

my $readBuffer;
my $BufferSize;
my $bytesRead;

my $bfOffBits;
my $biSize;
my $biWidth;
my $biHeight;
my $biBitCount;
my $biCompression;



if($num_args < 1){
  print "Call without parameters, exiting.\n";
  exit(1);
}

$SourceFile = substr($ARGV[0], 0, index($ARGV[$i], ".bmp"));

for($color = 0; $color < $color_no; $color++){

  open my $fh, '<:raw', $SourceFile . ".bmp" || die "Can't open $SourceFile.bmp!\n";
    $bytesRead = read $fh, $stringRead, 2;

    print "stringRead=$stringRead\n";
    if(($stringRead ne "BM")){
      print "Identifier $stringRead (BM expected). File is not a bitmap file!\n";
    }
    $bytesRead = read $fh, $readBuffer, 2*4;

    $bytesRead = read $fh, $readBuffer, 4;
    $bfOffBits = unpack 'L', $readBuffer;
    print "Header offset = $bfOffBits\n";

    $bytesRead = read $fh, $readBuffer, 4;
    $biSize = unpack 'L', $readBuffer;
    print "Info header size = $biSize\n";

    $bytesRead = read $fh, $readBuffer, 4;
    $biWidth = unpack 'L', $readBuffer;
    print "Bitmap width = $biWidth\n";

    $bytesRead = read $fh, $readBuffer, 4;
    $biHeight = unpack 'L', $readBuffer;
    print "Bitmap height = $biHeight\n";

    $bytesRead = read $fh, $readBuffer, 2;

    $bytesRead = read $fh, $readBuffer, 2;
    $biBitCount = unpack 'S', $readBuffer;
    if($biBitCount == 24){
      print "Color depth = $biBitCount\n";
    }
    else{
      print "Color depth = $biBitCount, but 24 expected! Exiting...\n";
    }

    $bytesRead = read $fh, $readBuffer, 4;
    $biCompression = unpack 'L', $readBuffer;
    if($biCompression == 0){
      print "Compression type = $biCompression\n";
    }
    else{
      print "Compression type = $biCompression, but 0 expected! Exiting...\n";
    }

    $bytesRead = read $fh, $readBuffer, 5*4;

    if($bfOffBits > 54){#Discard rest of info header
      $bytesRead = read $fh, $readBuffer, $bfOffBits - 54;
    }

    my $imageLine;
    my $fillByte = ($biWidth * 3) % 4;
    my $i;
    my $j;

    for($i=0; $i<4; $i++){
      if(($biWidth * 3 + $i) % 4 == 0){
        $fillByte = $i;
      }
    }

    for($j=0; $j<$biHeight; $j++){
      $imageLine = "";
      for($i=0; $i<$biWidth; $i++){
        $bytesRead = read $fh, $readBuffer, 3;
        ($rgb_b, $rgb_g, $rgb_r) = unpack 'C C C', $readBuffer;

        if($rgb_r < 128){
          $rgb_r = 0;
        }
        else{
          $rgb_r = 255;
        }

        if($rgb_g < 128){
          $rgb_g = 0;
        }
        else{
          $rgb_g = 255;
        }

        if($rgb_b < 128){
          $rgb_b = 0;
        }
        else{
          $rgb_b = 255;
        }

        $color_new = 0;#Pen 0 is default color
        if($rgb_r == 0 && $rgb_g == 0 && $rgb_b == 0){
          $color_new = 0;
        }

        if($rgb_r == 255 && $rgb_g == 0 && $rgb_b == 0){
          $color_new = 1;
        }

        if($rgb_r == 0 && $rgb_g == 255 && $rgb_b == 0){
          $color_new = 2;
        }

        if($rgb_r == 0 && $rgb_g == 0 && $rgb_b == 255){
          $color_new = 3;
        }
        #White color, pen 99 is ignored (0-3)!
        if($rgb_r == 255 && $rgb_g == 255 && $rgb_b == 255){
          $color_new = 99;
        }

#        print"R=$rgb_r, G=$rgb_g, B=$rgb_b, color_new=$color_new\n";

        if($color_new == $color){
          if($color_old != $color){#Change pen
            if($color_old != -1){#Start plotting, no need to park a pen
              print"color_old=$color_old, color=$color\n";
              &ParkPen($color_old);
            }
            &PickPen($color);
            $color_old = $color;
          }
          if($pen_mode){
            &SetDotFast($i, $j);
          }
          else{
            &SetDot($i, $j);
          }
        }

      }
      if($fillByte > 0){
        $bytesRead = read $fh, $readBuffer, $fillByte;
      }
      #print $imageLine . "\n";
    }

  close($fh);
  print "Number of dots (color $color) = " . $dot_no . "\n";
}#for($color = 0; $color < $color_no; $color++)

if($color_old != -1){#finally park pen
  &ParkPen($color_old)
}

print $coord_no . " coordinates found. Processing data...\n";
open(FILEOUT,'> ' . $SourceFile . '.gcode') || die "Can not open $SourceFile.gcode!\n";
  #print FILEOUT "M3 ;Spindle ON"
  #print FILEOUT "M8 ;Coolant ON"

  #needed to center plot on build plate
  $x_offset = ($bed_size_x-($x_max-$x_min))/2;
  $y_offset = ($bed_size_y-($y_max-$y_min))/2;

  print FILEOUT "G21 ; set units to millimeters\n";
  print FILEOUT "G90 ; use absolute coordinates\n";
  print FILEOUT "G1 Z" . sprintf("%.3f", $Z) . " ; Lift pen\n";
  print FILEOUT "G28 X0 Y0 ; home X and Y axis\n";
  print FILEOUT "G28 Z0 ; home Z axis\n";
  print FILEOUT "M302 P1 ; Prepare printer for cold extrusion\n";
  print FILEOUT "M203 X200 Y200 Z30 ; Set max feedrates\n";#MARLIN version (M203), ignored by Repetier

  #Start at 0/0 z_up
  $X = 0;
  $Y = 0;
  $Z = $Z_up;
  $X =  $x_offset;
  $Y =  $y_offset;
  $Z =  $Z;

  print FILEOUT "G1 Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . " ; Move Pen to UP position\n";
  print FILEOUT "G1 E" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . " ; Move Pen to UP position\n";
  print FILEOUT "G1 X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . " ; Move pen to 0/0\n";


  if($x_offset < 0){
    print "Width of file too large for print bed!\n";
  }
  if($y_offset < 0){
    print "Height of file too large for print bed!\n";
  }


  my $X_old = -999.999;
  my $Y_old = -999.999;
  my $Z_old = -999.999;
  my $E_old = -999.999;
  my $F_old = -999.999;
  for($i = 0; $i < $coord_no; $i++){
    if(index($remarks[$i], "keep X")>-1){
      $X = $X;#Use previous X coordinate
    }
    else{
      if(index($remarks[$i], "PenAction:")==-1){
        if($mirror_X == 0){
          $X = $X_coords[$i] - $x_min + $x_offset + $pen_X;
        }
        else{
          $X = ($x_max - $X_coords[$i]) + $x_offset + $pen_X;
        }
      }
      else{
        $X = $X_coords[$i];
      }
    }

    if(index($remarks[$i], "keep Y")>-1){
      $Y = $Y;#Use previous Y coordinate
    }
    else{
      if(index($remarks[$i], "PenAction:")==-1){
        if($mirror_Y == 0){
          $Y = ($y_max - $Y_coords[$i]) + $y_offset + $pen_Y;
        }
        else{
          $Y = $Y_coords[$i] - $y_min + $y_offset + $pen_Y;
        }
      }
      else{
        $Y = $Y_coords[$i];
      }
    }

    $Z =  $Z_coords[$i];
    if($X != $X_old || $Y != $Y_old || $Z != $Z_old || $feedrate_array[$i] != $F_old){
      print FILEOUT "G1 X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feedrate_array[$i]);
      if(length($remarks[$i])>0){
        print FILEOUT "; " . $remarks[$i];
      }
      print FILEOUT "\n";
      $X_old = $X;
      $Y_old = $Y;
      $Z_old = $Z;
      $F_old = $feedrate_array[$i];
    }

    if($E_coords[$i] != $E_old){
      print FILEOUT "G1 E" . sprintf("%.3f", $E_coords[$i])  . " F" . sprintf("%.3f", $feedrate_filament);
      if(length($remarks[$i])>0){
        print FILEOUT "; " . $remarks[$i];
      }
      print FILEOUT "\n";

      $E_old = $E_coords[$i];
    }
  }

  $X = 0;
  $Y = 0;
  $Z = $Z_up;
  print FILEOUT "X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . "\n";
  #print FILEOUT "M3\n";
  #print FILEOUT "M8\n";

  #print FILEOUT "M5 ;Spindle OFF"
  #print FILEOUT "M9 ;Coolant OFF"

  print FILEOUT "G1 Z" . sprintf("%.3f", $Z_up) . " ; Lift pen\n";
  print FILEOUT "G28 X0 ; home X axis\n";
  print FILEOUT "M84 ; disable all stepper motors\n";


  print "x_max = $x_max, x_min = $x_min, ymax = $y_max, y_min = $y_min \n";
  print "width = " . ($x_max - $x_min) . ", height = " . ($y_max- $y_min) . "\n";
  print "X offset = " . ($x_offset) . ", Y offset = " . ($y_offset) . "\n";
  print "Number of dots = " . $dot_no . "\n";

  print "...done!\n";

close(FILEOUT);

if(($x_max - $x_min) > $bed_size_x){
  print "\n\n########################################################\n";
  print "Width of plot larger than width of print bed !!!!\n";
  print "Change value of \$scale and try again!\n";
  print "Plotting this file might cause damage to your device!!!!\n";
  print "########################################################\n\n";
}
if(($y_max - $y_min) > $bed_size_y){
  print "\n\n########################################################\n";
  print "Height of plot larger than height of print bed !!!!\n";
  print "Change value of \$scale and try again!\n";
  print "Plotting this file might cause damage to your device!!!!\n";
  print "########################################################\n\n";
}




#+++++++++++++++++++++++++ Set dot ++++++++++++++++++++++++++++++++++++
sub SetDot{

  $dot_no++;

  $coord_no++;
  push @remarks, "keep X keep Y keep E";
  push @feedrate_array, $feed_rate;
  push @X_coords, -999999.999999;
  push @Y_coords, -999999.999999;
  push @Z_coords, $Z_up;
  push @E_coords, 0.000;


  $coord_no++;
  push @remarks, "";
  push @feedrate_array, $feed_rate;
  push @X_coords, $_[0] * $dotWidth;
  push @Y_coords, $_[1] * $dotHeight;
  push @Z_coords, $Z_up;
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "";
  push @feedrate_array, $feed_rate;
  push @X_coords, $_[0] * $dotWidth;
  push @Y_coords, $_[1] * $dotHeight;
  push @Z_coords, $Z_down;
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "";
  push @feedrate_array, $feed_rate;
  push @X_coords, $_[0] * $dotWidth;
  push @Y_coords, $_[1] * $dotHeight;
  push @Z_coords, $Z_up;
  push @E_coords, 0.000;


  if($_[0] * $dotWidth > $x_max){
     $x_max = $_[0] * $dotWidth;
  }
  if($_[0] * $dotWidth < $x_min){
    $x_min = $_[0] * $dotWidth;
  }
  if($_[1] * $dotHeight > $y_max){
    $y_max = $_[1] * $dotHeight;
  }
  if($_[1] * $dotHeight < $y_min){
    $y_min = $_[1] * $dotHeight;
  }

}

#+++++++++++++++++++ Set dot fast (Extruder motor) +++++++++++++++++++++
sub SetDotFast{

  $dot_no++;

  $coord_no++;
  push @remarks, "keep X keep Y";
  push @feedrate_array, $feed_rate;
  push @X_coords, -999999.999999;
  push @Y_coords, -999999.999999;
  push @Z_coords, $Z_down;
  push @E_coords, $E_up;


  $coord_no++;
  push @remarks, "";
  push @feedrate_array, $feed_rate;
  push @X_coords, $_[0] * $dotWidth;
  push @Y_coords, $_[1] * $dotHeight;
  push @Z_coords, $Z_down;
  push @E_coords, $E_up;

  $coord_no++;
  push @remarks, "";
  push @feedrate_array, $feed_rate;
  push @X_coords, $_[0] * $dotWidth;
  push @Y_coords, $_[1] * $dotHeight;
  push @Z_coords, $Z_down;
  push @E_coords, $E_down;

  $coord_no++;
  push @remarks, "";
  push @feedrate_array, $feed_rate;
  push @X_coords, $_[0] * $dotWidth;
  push @Y_coords, $_[1] * $dotHeight;
  push @Z_coords, $Z_down;
  push @E_coords, $E_up;

  if($_[0] * $dotWidth > $x_max){
     $x_max = $_[0] * $dotWidth;
  }
  if($_[0] * $dotWidth < $x_min){
    $x_min = $_[0] * $dotWidth;
  }
  if($_[1] * $dotHeight > $y_max){
    $y_max = $_[1] * $dotHeight;
  }
  if($_[1] * $dotHeight < $y_min){
    $y_min = $_[1] * $dotHeight;
  }

}

#+++++++++++++++++++++++++ Park Pen ++++++++++++++++++++++++++++++++++++
sub ParkPen{

  print"Parking pen...\n";

  $coord_no++;
  push @remarks, "Parking pen $_[0] (keep X keep Y). Going to low pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, -999999.999999;
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Parking pen $_[0] (keep Y). ...X pos gap...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Parking pen $_[0] (keep Y). ...high pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_high[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Parking pen $_[0] (keep Y). ...X pos parking...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_high[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Parking pen $_[0] (keep Y). ...low pos parking done!";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];
  push @E_coords, 0.000;

  print"...parked!\n";
}


#+++++++++++++++++++++++++ Pick Pen ++++++++++++++++++++++++++++++++++++
sub PickPen{

  $coord_no++;
  push @remarks, "Picking pen $_[0] (keep X keep Y). Going to low pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, -999999.999999;
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...X pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...X pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low_empty[$_[0]];
  push @E_coords, 0.000;

  my $amplitude = 0.5;
  my $i;
  my $sign = 0;
  for($i=$pen_Z_low_empty[$_[0]]; $i<$pen_Z_high[$_[0]]; $i+=$amplitude*2){
    $coord_no++;
    push @remarks, "PenAction: Pen pen $_[0] (keep Y). ...shaking up...";
    push @feedrate_array, $feed_rate_pen_change;
    if($sign==0){
      push @X_coords, $pen_X[$_[0]]+$amplitude;
      $sign=1;
    }
    else{
      push @X_coords, $pen_X[$_[0]]-$amplitude;
      $sign=0;
    }
    push @Y_coords, -999999.999999;
    push @Z_coords, $i;
    push @E_coords, 0.000;
  }

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...high pos...";
  push @feedrate_array, $feed_rate_pen_change;

  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_high[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...X pos gap high...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_high[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...X pos gap low, done!";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];
  push @E_coords, 0.000;

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y) to HIGH pos Extruder mode, done!";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];
  $E = $E_up;
  push @E_coords, $E;
}
